\ 3iForth.fs -- 3-Instruction Forth for AVR Atmel 328 / Arduino Nano
\ Assad Ebrahim, assadebrahim2000@gmail.com, http://www.mathscitech.org/articles
\ 2020-06-22

\ See Article & Discussion at http://www.mathscitech.org/articles/3iForth
\ TARGET = 3iForth.avr running on ATmel328 in Arduino Nano (compiled & uploaded to Nano using free Arduino IDE & default bootloader)
\ HOST = 3iForth.fs running in GForth 0.7.0 on Windows PC
\ Nano on startup (or reset) sets up serial port (9600 baud), initializes data stack, and waits for instructions over serial line

\ c. 150 lines of Forth code

\ USAGE:  
\   1. Connect Arduino Nano to HOST using USB mini cable
\           PINS - note: A0 through A7 are pins 14 to 21.  A0 to A5 (14-19) are I/O, A6/A7 (20-21) are AnalogRead only.
\   2. Run RealTerm with command line setup: baud=9600 half port=6.  This sets up the serial ports correctly to use 
\   3. HOST > run Gforth > include 3iForth.fs
\   4. HOST > help, help1, help2   (provides instructions)
\ Demo board 1: LED board with 11 LEDs alternating red and yellow (pins 2-12) and a tiny speaker (pin 13)
\ Demo board 2: Programmable Digital Synthesizer board with 13 keys and speaker

( * SERIAL PORT - these words open and close a given COM port on the PC * )
(   ex:            5 open-com              )
(   MISSING:  need to add GForth code for configuring the COM port.  Workaround: Use RealTerm to connect first  )

variable COMhandle
    : save-handle ( handle -- ) COMhandle ! ;
    : fh ( -- handle ) COMhandle @ ;

: com-string ( port# -- c-adr cnt )
    \ Builds COM port string.  Test: 6 com-string type
    >r s" comX" 2dup + 1 - '0' r> + swap c! ; 
: com-open? ( handle errcode -- )
    dup 0= ." ..." 
    if drop save-handle ." COM port opened OK." cr 
    else cr ." **error: check device is plugged into correct COM port" cr throw
    then ;
:  [OPEN-COM] ( port# --  )         
    \ Opens COM port in binary mode.
    \ Settling time delay of 1000ms doesn't work, 1500ms works
    com-string r/w bin create-file com-open? 1500 ms ;  
:  CLOSE-COM ( -- )
    fh close-file throw cr ." COM port closed." cr ;


( * MEMORY  - these words are essential for working efficiently with memory buffers, i.e. arrays of memory * )

: bytes  1 * ;
: words  2 * ;   
: dwords 4 * ;   \ GForth is 32-bits
\ Efficient fetch & store from/into and an array
\ BYTES
:  c@+ ( adr -- adr' val )  
    dup 1 bytes + swap    c@ ;
:  c!+ ( val adr -- adr' )  
    dup 1 bytes + -rot    c! ;
\ WORDS
: bytes2word ( hi8 lo8 -- val16 )
	swap    8 lshift   or ;
: lohi ( val16 -- hi8 lo8  )  \ little endian - low on top of stack.  ex: $FFFE lohi
    dup   $FF00 and   8 rshift    swap $FF and ;
    \ 440 f2hl 1 184
: w@ ( adr -- val16 )  \ little endian retrieval
	c@+    swap c@    swap  bytes2word ;
: w@+  ( adr -- adr' val16 )  
	c@+    swap c@+   swap -rot  swap  bytes2word ;
: w!  ( val16 adr -- )  \ little endian storage
	swap lohi rot  c!+ c! ;
: w!+ ( val16 adr -- adr' )  
	swap lohi rot  c!+ c!+ ;
\ DWORDS
: @+ 
    dup 1 dwords + swap @ ;
: !+
    dup 1 dwords + -rot ! ;


( * SERIAL READ - get messages from target into receive buffer * )

create RCV-BUFFER 255 bytes allot  \ buffer for holding messages from the uC
: show-rcv ( cnt -- cnt )   \  ex: 255 show-rcv
    \ Prints rcv buffer in decimal  ( use hex. instead of . to show values in hex  )
    dup RCV-BUFFER swap dup 0> 
    if 0 cr ." <rcv: " do c@+ . ." " loop else drop then drop ;    
: read-com ( n -- cnt )   \ read N bytes  ex: 50 read-com show-rcv
    RCV-BUFFER swap fh read-file throw ( -- cnt ) ;
: bytes. ( cnt -- cnt ) dup ." >> " . ." bytes on port" ; 

\ COMPLICATION: for REAL-TIME read, i.e. after sending a serial command and expecting a response from the uC on Serial.  80 ms works, 60 intermittent, 25 doesn't work
: rxN ( N -- cnt ) 80 ms read-com bytes. show-rcv ;     
: rx ( -- ) 50 rxN drop ;
: open-com ( port# -- ) [open-com] 5 rxN drop ;

( * SERIAL TRANSMIT - send messages in prepared message buffer to target * )

create MSG-BUFFER 20 bytes allot     \  write buffer
: msg-len ( adr' -- n )  
    MSG-BUFFER - ;
: show-msg ( cnt -- )                \ ex: 10 show-msg
    MSG-BUFFER swap 0 cr ." MSG-BUFFER>: " do c@+ . ." " loop drop ;          \ change . to hex. to show hex values
: [debug-msg] ( msglen -- ) dup show-msg ;
: tx  ( adr'  -- )                      
    MSG-BUFFER swap
    msg-len   \  [debug-msg]  
    fh write-file throw fh flush-file throw  ; \ transmit

\ TESTING ON PIN13
: test1-tx ( -- )  0 6 13 1 1 1 MSG-BUFFER c!+ c!+ c!+ c!+ c!+ c!+   6 show-msg tx ;  \ 13 out
: test2-tx ( -- )  1 6 13 1 1 1 MSG-BUFFER c!+ c!+ c!+ c!+ c!+ c!+   6 show-msg tx ;  \ 13 high
: test3-tx ( -- )  1 6 13 1 0 1 MSG-BUFFER c!+ c!+ c!+ c!+ c!+ c!+   6 show-msg tx ;  \ 13 low
: test-tx ( -- ) test1-tx test2-tx ;


( * SENDING REMOTE INSTRUCTIONS TO TARGET - some work with parameters that must reside on the target's data-stack * )
\ Notes: 
\ 1.   tx ;      only xcall eventually transits.  the others prepare the MSG-BUFFER (fewer comms -> more efficient)
\ 2.   -- adr'   each function accepts an address and provides the next address for easy chaining into MSG-BUFFER

\ EXECUTE 
1 constant INSTR_XSTORE
2 constant INSTR_XFETCH
3 constant INSTR_XCALL
4 constant INSTR_X_S
2 constant INSTR_XDROP

: [newmsg] ( -- adr' ) MSG-BUFFER ;
: [store]  ( a adr -- adr' ) INSTR_XSTORE swap c!+ c!+ ;
: [store2] ( a b adr -- adr' )  >r swap r> [store] [store] ;
: [store3] ( a b c adr -- adr' ) swap >r [store2] r> swap [store] ;
: [store4] ( a b c d adr -- adr' ) swap >r [store3] r> swap [store] ;
\ : [store4] ( a b c d adr -- adr' ) -rot >r >r [store2] r> r> rot [store2] ;
: [store5] ( a b c d e adr -- adr' ) swap >r [store4] r> swap [store] ; 

: x_s        ( -- )          INSTR_X_S    [newmsg] c!+ tx rx ;
: xstore     ( a -- )        [newmsg] [store] tx ;
: xfetch     ( -- val ) 	 INSTR_XFETCH [newmsg] c!+ tx 1 read-com rcv-buffer c@ ; 
: xcall      ( adr tok -- )  INSTR_XCALL rot c!+ c!+ tx ;
: xdrop      ( -- )          INSTR_XDROP  [newmsg] c!+ tx ;

( * HOST INTERFACE  CALL - on the host machine we don't need to know the mechanism, we think we live in the target and call the exported functions directly using Forth * )
\ Note: stack diagram is the same for .fs and for .avr.  This is because the [store] prep function reverses the entries

\ WARNING: these match the jumptable entries in the chip
0 constant INSTR_PINMODE
1 constant INSTR_PINSET
2 constant INSTR_DELAY
3 constant INSTR_TONE
4 constant INSTR_SETREG
5 constant INSTR_SWITCHKEY
6 constant INSTR_RECORD
7 constant INSTR_READPIN
8 constant INSTR_KEYPAD

1 constant OUTPUT
0 constant INPUT
2 constant INPUT_PULLUP
\ NO built-in Arduina constant for INPUT_PULLDOWN

: pinMode ( mode pin -- )    \ ex: 1 9 pinmode  (output)  modes: 0x0 INPUT, 0x1 OUTPUT, 0x2 INPUT_PULLUP
    [newmsg] [store2] INSTR_PINMODE xcall 60 ms ; \ Note: need settle time on pin when changing mode.  60 ms is magic number
: out ( pin -- )   OUTPUT swap pinmode ;
: in ( pin -- )    INPUT swap pinmode ;
: pullup ( pin -- ) INPUT_PULLUP swap pinmode ;  \ ex: 2 pullup
: pullup-keys  \ set all keyboard pins 2-12 to INPUT_PULLUP
    13 2 do i pullup loop 
    \ NEED TO ALSO DO A0 and A1 -- not sure how to pass those parameters!
    100 ms ; 

1 constant HIGH_SIGNAL
0 constant LOW_SIGNAL

: pinset ( level pin -- )  \ ex: 1 9 pinset   0 9 pinset
    [newmsg] [store2] INSTR_PINSET xcall ;
: digitalWrite pinset ;
: high ( pin -- )  HIGH_SIGNAL swap pinset ;
: low ( pin -- )   LOW_SIGNAL swap pinset ;
: lo low ;
: hi high ;
: testpin ( n -- )  dup dup out high 500 ms low ." done." ;

: delayms ( ms -- )   \ ex: 128 delayms   NEEDED on TARGET to ensure timings are preserved when running untethered
   MSG-BUFFER [store] INSTR_DELAY xcall ; 
: delay delayms ;

: [tone] ( freq-hi freq-lo dur-hi dur-lo pin -- )
    [newmsg] [store5] INSTR_TONE xcall ;
   \ 13 setspeaker
   \ 440 lohi 200 lohi speaker [tone]
   \ https://www.arduino.cc/reference/en/language/functions/advanced-io/tone/
   \ https://github.com/bhagman/Tone#ugly-details

: setreg ( val16 reg -- )   \ ex: 440 0 setreg    sets register0 to musical note A4
    swap lohi rot [newmsg] [store3] INSTR_SETREG xcall ;

: switchkey ( -- )  \ switch hardware modes
    [newmsg] INSTR_SWITCHKEY xcall ;

: record  ( on1/off0 -- )   \ turn on/off recording
	[newmsg] [store] INSTR_RECORD xcall ;

: [sendread] ( pin -- ) 
	[newmsg] [store] INSTR_READPIN xcall ;

: readpin ( pin -- 0/1 )   \ digitalRead() - with realtime delay (COMPLICATION)
	[sendread] 
	60 ms   \ 50 ms seems to work ( 0 failures out of 30 trials )
	1 read-com if rcv-buffer c@ else cr ." FAIL (readpin): no bytes" cr then ;

: pollkeypad ( -- val )  \ read matrix membrane keypad and send back 0x0-0xF numbers
	[newmsg] INSTR_KEYPAD xcall
	200 ms
	1 read-com if rcv-buffer c@ else $FF then ;

( * PROBLEM ORIENTED LANGUAGE - ON THE HOST --- setting up a simplified yet powerful custom dictionary * )

: reload ( -- )   \ ensure com port is closed before reloading which clobbers fh
  fh 0> if close-com then
  s" 3iForth.fs" included ;

( * HELP FILES * )

: help
    cr cr cr cr cr
    cr ."  ------------------------------------------------------------"
    cr ."  3-instruction Forth for Atmel 328 chip:"
    cr ."     STARTING/STOPPING"
    cr ."       5 open-com    -  opens communication to port 5 "
    cr ."       close-com "
    cr ."       reload "
    cr ."     OUTPUT PINS"
	cr ." 		2 out         - sets pin 2 into pullup high mode"
    cr ."       2 high        - sends voltage to pin 2  (choose pins 2 through 12)"
    cr ."       2 low         - sends voltage to pin 2  (choose pins 2 through 12)"
	cr ."	 INPUT PINS"
    cr ."       2 pullup     - sets pin 2 into pullup high mode"
	cr ."       2 readpin

    cr ."   => LEDs & BUZZERS:  include 3iLight.fs"
    cr ."       all-out  all-high all-low  5 n-vivid
    cr ."   => SOUND / MUSIC (tone)   include 3iSound.fs"  
    cr ."       start-speaker  - sets pin 13 to speaker" 
	cr ."    LOWEST LEVEL => help1"
    cr ."       host:     1 xstore     xdrop    x_s      [newmsg] 0 xcall   "
    cr ."       target:   pinmode   pinset   ms   tone   setreg   switchkey  "
    cr cr
;

: help1 ( -- )
    cr ."  2.  1  13 pinmode       In general:  pinMode ( mode pin --  )"
    cr ."  2.  2   9 pinmode       set pin 9 to INPUT_PULLUP
    cr ."  3.  1  13 pinset        In general:  pinset  ( level pin -- )   (also called digitalWrite() in Arduino C)."
    cr ."  4.  0  13 pinset        "
;

\ ======================
help
