( Sounds Program for 3-instruction Forth on Atmel 328 Arduino Nano )

( * WORDS FOR COMPUTERIZED MUSIC * )

\ Design: music is a time series of NOTES (frequencies), their DURATIONS (ms), and the RESTS (ms) between them.

\ Simplest play (v1) is specifying the frequencies and using fixed duration (120ms) and fixed rest (120-240ms).

\ Better (v2) is three arrays holding notes frequencies, note durations, and rests.  Play interleaves the instrument with all three.


variable SPEAKERpin  \ note - this is speaker for the PROGRAMMABLE SYNTHESIZER.  AVR 3iForth sets a DIGITAL PIANO speaker.  
                     \ If the two are different, then can have two speakers and voices...
    : setspeaker ( pin -- ) SPEAKERpin ! ;
    : speaker ( -- pin ) SPEAKERpin @ ;

variable dur_ms 120 dur_ms ! \ basic beat duration ms  (sixteenth)
    : beat dur_ms @ ; 
    : setbeat ( n -- ) dur_ms ! ;
    
: R1 beat ms ;
: R2 beat 2 * ms ; \ rest
: tone ( freq -- )
    lohi beat lohi speaker [tone] R2 ;
    \ 13 setspeaker
    \ 440 tone


( * MUSICAL NOTES   --   provides frequency and generates tone * )
\ Calculation for the half steps:  
\ The mathematics of western music is multiplicative (equitempered scale popularized by Bach in the 1700s)
\ The musical scale has 12 semitones with each octave being double the frequency of the previous
\ So the ratio between successive musical semitones is fixed as 12th-root of 2 = 2^(1/12) = 1.05946309436
\ By convention, there are 2 flats (b%, e%) and three sharps (c#, f#, g#)
\ 73 semitones covering 6 octaves (=12*6+1)
create piano   \ 108 semitones, spanning 9 octaves, 
    \       C     d%     D      e%     E      F      F#     G      a%     A      b%     B
    \             1             3                    6             8             10
    \       0            2             4      5             7             9            11      
\ ( 0 )      16 ,   17 ,   18 ,   19 ,   21 ,   22 ,   23 ,   25 ,   26 ,   28 ,   29 ,   31 , 
\ For Arduino Atmel 328 - It is not possible to generate tones lower than 31Hz (B0)
( 0 )      33 ,   35 ,   37 ,   39 ,   41 ,   44 ,   46 ,   49 ,   52 ,   55 ,   58 ,   62 ,
( 1 )      65 ,   69 ,   73 ,   78 ,   82 ,   87 ,   93 ,   98 ,  104 ,  110 ,  117 ,  124 ,
( 2 )     131 ,  139 ,  147 ,  156 ,  165 ,  175 ,  185 ,  196 ,  208 ,  220 ,  233 ,  247 ,
( 3 )     262 ,  277 ,  294 ,  311 ,  330 ,  349 ,  370 ,  392 ,  415 ,  440 ,  466 ,  494 ,  \ middle
( 4 )     523 ,  554 ,  587 ,  622 ,  659 ,  698 ,  740 ,  784 ,  831 ,  880 ,  932 ,  988 , 
( 5 )    1047 , 1109 , 1175 , 1245 , 1319 , 1397 , 1480 , 1568 , 1661 , 1760 , 1865 , 1976 , 
( 6 )    2093 , 2217 , 2349 , 2489 , 2637 , 2794 , 2960 , 3126 , 3322 , 3520 , 3729 , 3951 , 
( 7 )    4186 , 4435 , 4699 , 4978 , 5274 , 5588 , 5920 , 6272 , 6645 , 7040 , 7459 , 7908 ,
96 constant semitones  \ spanning 8 octaves

: f2i  ( freq -- idx )  \ find index given frequency
   >r piano begin @+ r@ = until r> drop
   piano - 4 / 1 - ;

\ Playing by the absolute numbers  ( 0 to 72 )
: freq ( idx -- )  \ ex: 12 freq tone 
    dwords piano + @ ;
\ Playing by 2-D coordinations
: 2di ( irow jcol -- idx )  \ 2d coordinates to 1d index
    swap 12 * + ;
\ Playing by root and offsets (relative numbers)
variable zero 
    : set-zero ( i j -- ) 2di zero c! ;  \ ex: 3 0 set-zero
    : reset-zero 0 0 set-zero ;
    reset-zero
    : show-zero  ( -- idx ) zero c@ ;

\ Playing music 
: note ( idx -- ) show-zero + freq tone ;
: note2d ( irow icol -- ) 2di note ;
\ Playing music by number relative to note-offset (show-zero)
: ., ( num1-8 -- ) 1 - note ;

: upward  semitones 0 do i note loop ;
: downward semitones 0 do semitones 1 - i - note loop ;
: updown upward downward ;
: updown-fast 60 setbeat updown 120 setbeat ;
: updown-sfast 30 setbeat updown 120 setbeat ;
: updown-slide 15 setbeat updown 120 setbeat ;
: updown-roar 8 setbeat updown 120 setbeat ; 

create white-registers 0 c, 2 c, 4 c, 5 c, 7 c, 9 c, 11 c, 12 c,
create black-registers 1 c, 3 c, 6 c, 8 c, 10 c, 

: xscale5 ( scale -- )  \ ex: cstd5 xscale5
    5 0 do dup i + c@ freq black-registers i + c@ setreg loop drop ;
: black-null  ( -- ) 5 0 do 0 black-registers i + c@ setreg loop ;
: xscale8 ( scale -- )  \ send scale to target  ex: cmaj8 xscale8
    8 0 do dup 
        i + c@ freq
        white-registers i + c@ 
        setreg 
    loop black-null drop ;

\ Playing by reference against a scale
variable scale-ptr    \ holds ptr to scale, allowing playing tunes against 8-key predefined scale
    : scale  ( -- a* )  scale-ptr @ ;
    : set-scale ( a* -- )  scale-ptr ! ;
    
create cstd8 36 c, 38 c, 40 c, 41 c, 43 c, 45 c, 47 c, 48 c, 
create cstd5 37 c, 39 c, 42 c, 44 c, 46 c,  \ equitemper 
    : cstd  cstd8 set-scale  cstd8 xscale8  cstd5 xscale5 ;
create cmaj8 31 c, 33 c, 36 c, 38 c, 40 c, 43 c, 45 c, 48 c, 
    : cmaj  cmaj8 set-scale  cmaj8 xscale8 ;
create bhairav8  31 c, 32 c, 35 c, 36 c, 38 c, 39 c, 42 c, 43 c,
	: bhairav bhairav8 set-scale bhairav8 xscale8 ;
	


: .. ( n -- ) \ plays note from scale array
    1 - scale + c@ note ;
: play-scale  9 1 do i .. loop ;


: start-speaker ( programmable digital synthesizer )
    13 out
    13 setspeaker
    reset-zero
	cstd play-scale
    \ b. d. g. 
    ;


\ ETHNIC SYNTHESIZERS

\ setting up ethno tunes
: whites ( 1 2 3 4 5 6 7 8 -- )  \ setup white frequency-key mappings on target piano
    12 setreg 11 setreg 9 setreg 7 setreg 5 setreg 4 setreg 2 setreg 0 setreg ;
: blacks ( 1 2 3 4 5 -- )
    10 setreg 8 setreg 6 setreg 3 setreg 1 setreg ;
: null-black-keys ( -- )
    0 0 0 0 0 blacks ;

: init-keys
   10 13 setreg  \ set min duration (ms) of note
   5 14 setreg  \ set settle time following note
;
: equitemper  \ Bach's equitemper scale, do-re-me tuning
   262  294   330   349   392   440   494   523   whites
   277  311   370   415   466   blacks  ;
   
: set-cmaj  \ starts on G  She'll be coming round the mountain when she comes...   O! Susanna (start on 3)
    196 220 262 294 330 392 440 523
    whites null-black-keys ;

\ INDIAN SCALES
\ Source: Special Topics in Music Theory, by Catherine Schmidt-Jones
\    Ch. 3.5 Indian Classical Music: Tuning and Ragas
\ https://cnx.org/contents/mviDomhA@5.66:bDLyHWWs@16/Indian-Classical-Music-Tuning-and-Ragas
\ A "that" is a scale of 7 notes -- why 7?  (7 celestial bodies - Sun, Moon, Mercury, Venus, Mars, Jupiter and Saturn)
\ Aside: 7 days a week: Babylonians of the 6th century... https://www.discovermagazine.com/planet-earth/why-are-there-seven-days-in-a-week
\ A raga chooses five, six, or all 7 notes of the that, and sticks to that, i.e. pentatonic raga, hexatonic raga, or full raga

: bhairav.  \ raga with two flats on D and A    Sa Re Ga Ma Pa Dha Ni
   262   277   330   349   392   415   494  523   whites  null-black-keys ;
: asavari  \ raga with three flats on E, A, B    Sa Re Ga Ma Pa Dha Ni
   262   294   311   349   392   415   466   523  whites  null-black-keys ;
: bhairavi  \ raga with four flats on D, E, A, B    Sa Re Ga Ma Pa Dha Ni
   262   277   311   349   392   415   466   523 whites  null-black-keys ;
: kafi  \ raga with two flats on E and B    Sa Re Ga Ma Pa Dha Ni   ex: unchaare kot 3rd note, aaj aanand 1st note, and scarborough fair
   262   294   311   349   392   440   466   523  whites  null-black-keys ;
: khamaj \ raga with 1 flat on B
   262   294   330   349   392   440   466   523  whites  null-black-keys ;
: kalyan \ raga with 1 sharp on F
   262    294    330    370    392    440    494    523  whites  null-black-keys ;
: purvi \ raga with 1 sharp on F and two flats on D and A  "back to life, back to reality"
   262    277     330    370    392    415    494    523  whites  null-black-keys ;
: todi  \ bhairavi raga (four flats on D, E, A, B) with additionally 1 sharp on F    Sa Re Ga Ma Pa Dha Ni
   262    277    311    370    392    415    466    523  whites  null-black-keys ;
: bilawal. ; \ raga with no flats or sharps, ex: abadoo and doh-a-deer

\  CHINESE SCALES

: yu  \ Yu scale on Huang Zhong base freq ( Why Don't you come back... and Fire on the Mountain)
   440    529    594    660    793    440 2 *   529 2 *   594 2 *  whites  null-black-keys ;
: yu-1  \ Yu scale on Huang Zhong base freq shifted down one ( Fire on the Mountain )
   793 1 rshift   440   529  594   660   793  440 2 *   529 2 *  whites  null-black-keys  ;
: shang  \ Shang scale on Da Cu  base freq ( Snoopy theme)
   495    557    669    743    892    495 2 *     557 2 *    669 2 * whites  null-black-keys ;
: gong  \ Gong scale on Zhong Lu base freq
   594    669    752    892    1003    594 2 *   669 2 *   752 2 *   whites null-black-keys ;
: jue  \ Jue scale on Yi Ze base freq 
   705 0    847 2    952 4    1129 5   1270 7   705 2 *    847 2 *    952 2 * ;
: zhi  \ Zhi scale on Wu Yi base freq 
   793    892    1071   1189   1338   793 2 *    892 2 *    1071 2 *  ;



\ Frequency lookup table - basic
\ Note double is an octave higher, and half is an octave lower -- CONSIDER using fixed point and calculating the initial table
\ Flats and sharps as + / - against base
: C...    65 tone ;  \ octave lower 2
: c...#   69 tone ;
: D...    73 tone ;
: e...%   78 tone ;
: E...    82 tone ;
: F...    87 tone ;
: f...#   93 tone ;
: G...    98 tone ;     
: g...#  104 tone ;
: A...   110 tone ;
: b...%  117 tone ;
: B...   124 tone ;
: C..    131 tone ;  \ octave lower
: c..#   139 tone ;
: D..    147 tone ;
: e..%   156 tone ;
: E..    165 tone ;
: F..    175 tone ;
: f..#   185 tone ;
: G..    196 tone ;     
: g..#   208 tone ;
: A..    220 tone ;       
: b..%   233 tone ;
: B..    247 tone ;
: C.     262 tone ; \ middle octave -- Mddle C
: c.#    277 tone ;
: D.     294 tone ; 
: e.%    311 tone ;
: E.     330 tone ;        \ tone(SPEAKER,330,QUARTER); delay(SHORT_DELAY); } 
: F.     349 tone ; 
: f.#    370 tone ;
: G.     392 tone ;
: g.#    415 tone ;    
: a.%    g.# ;
: A.     440 tone ;
: b.%    466 tone ; 
: B.     494 tone ;
: C'     523 tone ;  \ octave upper 1
: c'#    554 tone ;
: D'     587 tone ;
: e'%    622 tone ;
: E'     659 tone ;
: F'     698 tone ;
: f'#    740 tone ;
: G'     784 tone ;
: g'#    831 tone ;
: A'     880 tone ;
: b'%    932 tone ;
: B'     988 tone ;
: C''   1047 tone ;  \ octave upper 2
: c''#  1109 tone ;
: D''   1175 tone ;
: e''%  1245 tone ;
: E''   1319 tone ;
: F''   1397 tone ;
: f''#  1480 tone ;
: G''   1568 tone ;
: g''#  1661 tone ;
: A''   1760 tone ;
: a''#  1865 tone ;
: B''   1976 tone ;
: C'''  2093 tone ;  \ octave upper 3
: c'''# 2217 tone ;
: D'''  2349 tone ;
: e'''% 2489 tone ; 
: E'''  2637 tone ;
: F'''  2794 tone ;
: f'''# 2960 tone ;
: G'''  3126 tone ;
: g'''# 3322 tone ;
: A'''  3520 tone ;
: b'''% 3729 tone ;
: B'''  3951 tone ;  \ 12-bits of frequency goes up to B'''
: C'''' 4186 tone ;  \ octave upper 4

\  === REPLACED by full array of notes
\ * SIMPLE NUMBERS TO NOTES via an array of note executables *
\ create notes ' c. , ' d. , ' e. , ' f. , ' g. , ' a. , ' b. , ' c' , 
\ : .. ( n -- )  \ plays note from array
\  1 - dwords notes + @ execute ;



: help-music
    cr ." Get started:"
    cr ."     plug in digital piano"
    cr ."     use RealTerm to open the connection to the port"
    cr ." In GForth:"
    cr ."     include 3iForth.fs"
	cr ."     include 3iSound.fs"
    cr ."     5 open-com  start-speaker"
    cr ." Several ways to play:"
    cr ."    1.  c.  c.. c' c'' c''#  b.. b..%  a. "
    cr ."    2.  440 tone"
    cr ."    3.  45 note"
    cr ."    5.  46 .,   ., plays the semitones from zero point"
    cr ."    6.  3 0 set-zero    c.  0 note    1 .,   12 note  13 .,  a.  9 note  10 .,  reset-zero"
    cr ."    4.  cstd    1 .. 6 ..   .. plays the notes on the scale of 8 preferred keys"
    cr ."    7.  playlist
    cr ." More help:  help-record   help-playlist  help-tunings  
;

: help2 ( -- )
    cr ."  2.  13 out       - set pin 13 to out"
    cr ."  3.  2 high        - sends voltage to pin 2  (choose pins 2 through 12)"
    cr ."  4.  2 low         - sends voltage to pin 2 (choose pins 2 through 12)"
    cr ."  5.  13 setspeaker    - identifies the speaker"
    cr ."  6. 440 tone"
    cr ."  7.  1 .. c... c.. c. d. e. c' c'' f'# -  sends tone to pin 13 (speaker)  c. is middle, c.. is octave lower, c' is octave higher "
;


: help-tunings
  cr ." Western styles: equitemper  cstd  cmaj"
  cr ." Chinese styles (5): yu yu-1(*) shang gong jue zhi"
  cr ." Indian styles (8): bhairav, asavari, bhairavi, kafi, khamaj, kalyan, purvi, todi"
;


: help-misc
    cr ."           440 tone      - sends 440 Hz (A) note to speaker"
    cr ."    c. d. e. c' c'' c.. c... f'# a.% -  sends tone to pin 13 (speaker)  c. is middle, c.. is octave lower, c' is octave higher, # sharp, % flat
    cr ."    1 .. 2 ..  - convert numbered simple piano to notes"
    
    cr ."  playlist   - runs through wings, harp, ode, memory, daffodils, morning, stairway, freedom, rainbow, reason, scarborough "
    cr ."  compositions: Assad: yearning snake  Jasmine: tog (together) "
    cr ."  => SOUND / MUSIC (tone)   include 3iSounds.fs"  
    cr ."       start-speaker  - sets pin 13 to speaker" 
    cr ."       record         - toggles recording mode for the keys (turns digital piano serial transmit on/off)"
    cr ."       save-bar       - saves 5 notes"
    cr ."       play-notes     - plays all saved notes"
    cr ."       delete-note    - delete last saved note"

;

\ ****************************
help-music
